//---------------------------------------------------------------------------
//
//    floating point to wave conversion program
//
//    post-modulation processing for digital communications Tx file
//
//    Copyright (C) 2002  Barry D. Sanderson
//    Copyright (C) 2002  Carl A. Harcourt
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Mailing addresses for the authors:
//
//    Barry D. Sanderson
//    1725 N. Bolton Ave.
//    Indianapolis, IN 46218  USA
//
//    Carl A. Harcourt
//    4937 W. 900 S.
//    Milroy, IN 46156  USA
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Rev 24 Sep 2001 (BDS) - version 2.01, added detection of wrong number
//                            of bytes written to disk file
//
//    Rev 30 Aug 2002 (CAH) - version 2.00, for initial PC/DOS release
//                            Developed using MS Visual C++ 6.0
//                            as a 'Win32 Console Application', ie a DOS .exe program
//                            Requires a Win32 environment for proper DOS execution
//                            This version will also build and execute unmodified on a Linux system
//
//    Rev 23 Apr 2001 (BDS) - version 1.0.0 for initial release
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// 'C' include files

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

//---------------------------------------------------------------------------
//  application include files

#include "Wcommon.h"

//---------------------------------------------------------------------------
// defines

#define F2W_INPUT_VAR_TYPE  float
#define F2W_OUTPUT_VAR_TYPE signed short int

#define WAVHEADERLENGTH         44
#define WAVLENGTH1OFFSET        4
#define WAVLENGTH2OFFSET        40
#define WAVLENGTHDIFFERENCE     (WAVLENGTH2OFFSET - WAVLENGTH1OFFSET)

//---------------------------------------------------------------------------
// variable declarations

char    cSignonMessage[]    = "Floating Point To Wav Conversion Program";
char    cVersionMajor[]     = "R 02";
char    cVersionMinor[]     = "01";

struct stat     statb;
unsigned long   ulFileLength, ulFileBufferIndex, ulNumberOfElements, ulInputElementsRead;

struct stat     tmp_stat;
int             tmp_status;

int             statr;

FILE         *  FpInputFile;
FILE         *  FpOutputFile;
FILE         *  FpTmpFile;

unsigned char   mp_sd_filename[] = "mp_tmp.sd";
unsigned char   mp_cl_filename[] = "mp_tmp.cl";
unsigned char   mp_sf_filename[] = "mp_tmp.sf";

F2W_INPUT_VAR_TYPE  *  InputFileBuffer;
F2W_OUTPUT_VAR_TYPE *  OutputFileBuffer;

unsigned char   ucWavHeaderBuffer [WAVHEADERLENGTH] =
    {
    0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00,
    0x57, 0x41, 0x56, 0x45, 0x66, 0x6d, 0x74, 0x20,
    0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00,
    0x11, 0x2b, 0x00, 0x00, 0x22, 0x56, 0x00, 0x00,
    0x02, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61,
    0x00, 0x00, 0x00, 0x00
    };

unsigned int    iErrorNo;

char            cMsgNoError[]      = "Successful termination";
char            cMsgUsage[]        = "Usage:\n\n  flt2wav input_file output_file clipping_level scale_factor\n";
char            cMsgUnknownError[] = "Unknown error\n";

char            cMsg03[]           = "Couldn't open input file\n";
char            cMsg04[]           = "Couldn't open output file\n";
char            cMsg05[]           = "WRONG number of bytes written to disk file\n";
char            cMsg06[]           = "Error in reading body of input file\n";
char            cMsg07[]           = "\n";
char            cMsg08[]           = "Initial memory allocation failed\n";
char            cMsg09[]           = "\n";
char            cMsg10[]           = "\n";
char            cMsg11[]           = "\n";
char            cMsg12[]           = "\n";
char            cMsg13[]           = "stat failed\n";

char *          cpErrorMsgTbl[]    = {cMsgNoError, cMsgUsage,  cMsgUnknownError, cMsg03, cMsg04, cMsg05, cMsg06,
                                      cMsg07,  cMsg08, cMsg09,  cMsg10, cMsg11, cMsg12, cMsg13};

//---------------------------------------------------------------------------
// function declarations

//---------------------------------------------------------------------------
int main (int argc, char *argv[], char *env[])

    //  arv[1] is input file name
    //  arv[2] is output file name
    //  arv[3] is clipping level
    //  arv[4] is scale factor

{

    F2W_INPUT_VAR_TYPE      scale_factor;
    F2W_INPUT_VAR_TYPE      standard_deviation;
    F2W_INPUT_VAR_TYPE      clipping_level;

    double                  tmp_value;
    unsigned int            wav_bc1;
    unsigned int            wav_bc2;
    size_t            val_written;

    //-----------------------------------------------------------------------
    // print sign on message
    fprintf (stdout, "\n%s - %s.%s\n\n", cSignonMessage, cVersionMajor, cVersionMinor);

    //-----------------------------------------------------------------------
    standard_deviation = 0.0;
    clipping_level     = 0.0;
    scale_factor       = 0.0;

    //----------------------------------------------------------------------
    // get standard deviation, clipping level, and scale factor
    // from intermdiate files, if present

    FpTmpFile = NULL;
    FpTmpFile = fopen (mp_sd_filename, "r");             // open intermediate file for read
    if (FpTmpFile != NULL)
    {
        fscanf (FpTmpFile, "%e", &standard_deviation);   // read standard deviation from intermediate file
        fclose (FpTmpFile);                              // close intermediate file
    }

    FpTmpFile = NULL;
    FpTmpFile = fopen (mp_cl_filename, "r");             // open intermediate file for read
    if (FpTmpFile != NULL)
    {
        fscanf (FpTmpFile, "%e", &clipping_level);       // read clipping_level from intermediate file
        fclose (FpTmpFile);                              // close intermediate file
    }

    FpTmpFile = NULL;
    FpTmpFile = fopen (mp_sf_filename, "r");             // open intermediate file for read
    if (FpTmpFile != NULL)
    {
        fscanf (FpTmpFile, "%e", &scale_factor);         // read scale_factor from intermediate file
        fclose (FpTmpFile);                              // close intermediate file
    }

    iErrorNo = 8;

    //----------------------------------------------------------------------
    // did intermediate files exist?

    if ((clipping_level == 0.0) || (scale_factor == 0.0) || (standard_deviation == 0.0))
    {                                       // no
        if (argc < 5)                           // enough arguments?
        {                                           // no
            fprintf (stderr, cpErrorMsgTbl[1]);
            exit (1);                               // exit
        }
        else
        {
            //--------------------------------------------------------------
            // get input values from command line invocation
            clipping_level = (F2W_INPUT_VAR_TYPE) atof (argv[3]);
            scale_factor   = (F2W_INPUT_VAR_TYPE) atof (argv[4]);
        }
    }

    //----------------------------------------------------------------------
    // print input values

    printf ("The standard deviation is %7.4f\n", standard_deviation);
    printf ("The clipping level is %7.4f\n",     clipping_level);
    printf ("The scale factor is %7.4f\n",       scale_factor);

    //----------------------------------------------------------------------
    // get input file info

    statr = stat (argv[1], &statb);
    if (statr != 0)
    {
        fprintf (stderr, cpErrorMsgTbl[13]);
        exit (13);
    }

    ulFileLength       = (unsigned long) (statb.st_size);
    ulNumberOfElements = ulFileLength/(sizeof(F2W_INPUT_VAR_TYPE));

    //----------------------------------------------------------------------
    // create file buffers

    if ((InputFileBuffer = malloc((size_t) ((ulNumberOfElements)* sizeof(F2W_INPUT_VAR_TYPE)))) == NULL)
        goto exit_a;
    if ((OutputFileBuffer = malloc((size_t) ((ulNumberOfElements)* sizeof(F2W_OUTPUT_VAR_TYPE) ))) == NULL)
        goto exit_b;
    iErrorNo = 3;

    //----------------------------------------------------------------------
    // open input file for reading

    FpInputFile = NULL;
    FpInputFile = fopen (argv[1], "rb");
    if (FpInputFile==NULL)
    {
        fprintf (stderr,"Could not open input file %s\n", argv[1]);
        goto exit_c;
    }
    iErrorNo = 4;

    //----------------------------------------------------------------------
    // open output file for writing

    FpOutputFile = NULL;
    FpOutputFile = fopen (argv[2], "wb");
    if (FpOutputFile==NULL)
    {
        fprintf (stderr,"Could not open output file %s\n", argv[2]);
        goto exit_d;
    }

    //----------------------------------------------------------------------
    // read input file

    ulInputElementsRead = fread (&InputFileBuffer[0], sizeof(F2W_INPUT_VAR_TYPE), ulNumberOfElements, FpInputFile);
    if (ulInputElementsRead != ulNumberOfElements)
    {
        iErrorNo=6;
        goto exit_e;
    }

    //----------------------------------------------------------------------
    //  clip and scale input values

    for (ulFileBufferIndex=0; ulFileBufferIndex<ulNumberOfElements; ulFileBufferIndex++)
    {
        tmp_value = InputFileBuffer[ulFileBufferIndex];
        if (tmp_value > clipping_level)
            tmp_value = clipping_level;
        else
        {
            if(tmp_value<-clipping_level)
                tmp_value = -clipping_level;
        }
        OutputFileBuffer[ulFileBufferIndex] = (F2W_OUTPUT_VAR_TYPE)(scale_factor*tmp_value);
    }

    //----------------------------------------------------------------------
    // put proper lengths into ucWavHeaderBuffer

    wav_bc2 = (unsigned int) ((ulNumberOfElements) * sizeof(F2W_OUTPUT_VAR_TYPE));
    wav_bc1 = (unsigned int) ((ulNumberOfElements) * sizeof(F2W_OUTPUT_VAR_TYPE) + WAVLENGTHDIFFERENCE);

    //----------------------------------------------------------------------
    // write wav file header

    iErrorNo=5;
    val_written= fwrite (&ucWavHeaderBuffer[0], 1, WAVLENGTH1OFFSET, FpOutputFile);
    if (val_written != (size_t)(WAVLENGTH1OFFSET)) goto exit_e;
    val_written= fwrite (&wav_bc1, sizeof(unsigned int), 1, FpOutputFile);
    if (val_written != (size_t)(1)) goto exit_e;
    val_written= fwrite (&ucWavHeaderBuffer[WAVLENGTH1OFFSET + sizeof(unsigned int)], 1, WAVLENGTHDIFFERENCE-sizeof(unsigned int), FpOutputFile);
    if (val_written != (size_t)(WAVLENGTHDIFFERENCE-sizeof(unsigned int))) goto exit_e;
    val_written= fwrite (&wav_bc2, sizeof(unsigned int), 1, FpOutputFile);
    if (val_written != (size_t)(1)) goto exit_e;
    val_written= fwrite (&OutputFileBuffer[0], sizeof(F2W_OUTPUT_VAR_TYPE), ulNumberOfElements, FpOutputFile);
    if (val_written != (size_t)(ulNumberOfElements)) goto exit_e;

    iErrorNo=0;

    //----------------------------------------------------------------------
    // exeunt all

exit_e:
    fclose(FpOutputFile);

exit_d:
    fclose(FpInputFile);

exit_c:
    free(OutputFileBuffer);

exit_b:
    free(InputFileBuffer);

exit_a:
    if (iErrorNo != 0)
        fprintf(stderr, cpErrorMsgTbl[iErrorNo]);

    exit(iErrorNo);
}
