//---------------------------------------------------------------------------
//
//    Reed-Solomon code
//
//    Copyright (C) 2002  Barry D. Sanderson
//    Copyright (C) 2002  Carl A. Harcourt
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Mailing addresses for the authors:
//
//    Barry D. Sanderson
//    1725 N. Bolton Ave.
//    Indianapolis, IN 46218  USA
//
//    Carl A. Harcourt
//    4937 W. 900 S.
//    Milroy, IN 46156  USA
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
//    Rev 30 Aug 2002 (CAH) - version 2.0, for initial PC/DOS release
//                            Developed using MS Visual C++ 6.0
//                            as a 'Win32 Console Application', ie a DOS .exe program
//                            Requires a Win32 environment for proper DOS execution
//                            This version will also build and execute unmodified on a Linux system
//
//    Rev 25 Nov 2001 (BDS) - version 1.1.0, for Wyman initial public release
//
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// application include files

#include "Wcommon.h"
#include "code_def.h"
#include "rd_sl_v.h"
#include "rd_sl_p.h"


//---------------------------------------------------------------------------
// function declarations

//---------------------------------------------------------------------------
int    encode2 (int *extra, int *info)
{
    //    extra       points to array where redundancy symbols of code word will
    //                be stored,
    //    info        points to array where information symbols have been stored
    //
    //    returns:     0 if no errors were encountered
    //                -1 if an illegal symbol is found in the "info" array

    int     si, di;
    int     tmpi;

    for (si=0; si<(MAX_GF_EL_2-2*MAX_CORR_ERRORS); si++)
        if ((info[si]<0) || (info[si]>MAX_GF_EL_2))
            return -1;
    for (di=0; di<(2*MAX_CORR_ERRORS); di++)
    {
        tmpi = 0;
        for (si=0; si<(MAX_GF_EL_2-2*MAX_CORR_ERRORS); si++)
        {
            tmpi = gf9add[tmpi][gf9mul[info[si]][rgm2[di][si]]];
        }
        extra[di] = tmpi;
    }

    return 0;
}

//---------------------------------------------------------------------------
int    decode2 (int *codeword, int rho, int *erase)
{
    //    codeword    points to the codeword to be decoded.  This is in the
    //                form [r r r r i i i i], where the r's are the
    //                redundancy symbols, and the i's are the information symbols,
    //                both with least significant symbols first.
    //    rho         is the number of erasures
    //    erase       points to an array of erasure indexes.  Each index is
    //                in the range 0 through n-1 (i.e. 0 through 8)
    //
    //    returns     a positive number indicating how many symbols were corrected
    //             0 if no symbols were in error
    //            -1 if any codeword symbols are illegal
    //            -2 if rho > 4 (too many erasures)
    //            -3 if any erasure indexes are out of bounds
    //
    //    if return value >= 0, then *codeword contains the corrected
    //      values.

    int     si, di;
    int     V[MAX_GF_EL_2], S[MAX_GF_EL_2], lam[MAX_GF_EL_2], B[MAX_GF_EL_2],
            T[MAX_GF_EL_2], del[NUM_GF_EL_2], C[MAX_GF_EL_2], c[MAX_GF_EL_2];
    int     sh[MAX_GF_EL_2];
    int     r, L, tmpi, tmpj, tmpk, texp, ec;

    for (si=0; si<MAX_GF_EL_2; si++)
    if ((codeword[si] < 0) || (codeword[si]> MAX_GF_EL_2))
        return -1;
    if (rho > (2*MAX_CORR_ERRORS))
        return -2;
    if (rho > 0)
    {
        for (si=0; si< rho; si++)
            if ((erase[si] < 0) || (erase[si] > (MAX_GF_EL_2-1)))
                return -3;
    }
    L      = 0;
    r      = 0;
    B[0]   = 1;
    lam[0] = 1;
    for (si=1; si<MAX_GF_EL_2; si++)
        {B[si]=0; lam[si]=0;}

    // Transform received codeword
    for (di=0; di<MAX_GF_EL_2; di++)
    {
        tmpi = 0;
        for (si=0; si<MAX_GF_EL_2; si++)
            {tmpi = gf9add[tmpi][gf9mul[expa2[(si*di)%MAX_GF_EL_2]][codeword[si]]];}
        V[di] = tmpi;
        S[di] = V[di];
    }

    //    note that while doing transforms S indexes range from 0 through MAX_GF_EL_2-1, but
    //    when revising values of S the index rnages from 1 to MAX_GF_EL_2.  This is
    //    resolved by using 0 through MAX_GF_EL_2-1 while transforming, and using
    //    1 through MAX_GF_EL_2 (mod MAX_GF_EL_2) when revising values of S.

    while(r<rho)
    {
        r++;
        {
            for (si=MAX_GF_EL_2-1; si>0; si--)
                sh[si]= gf9mul[expa2[erase[r-1]]][lam[si-1]];
            sh[0] = 0;
            for (si=0; si<MAX_GF_EL_2; si++)
                {lam[si]= gf9add[lam[si]][gf9addi[sh[si]]]; B[si]= lam[si];}
            L++;
        }
    } // end of (r<rho)

    while(r < 2*MAX_CORR_ERRORS)
    {
        r++;
        // calculate del[r]
        tmpi = 0;
        for (si=0; si<=(r<MAX_GF_EL_2 ? r: MAX_GF_EL_2-1); si++)
            tmpi = gf9add[tmpi][gf9mul[lam[si]][S[(r-si)%MAX_GF_EL_2]]];
        del[r%MAX_GF_EL_2] = tmpi;
        while(del[r%MAX_GF_EL_2] < 0)
            del[r%MAX_GF_EL_2] += NUM_GF_EL_2;
        if (del[r] == 0)
        {
            for (si=MAX_GF_EL_2-1; si>0; si--)
                B[si] = B[si-1];
                B[0]  = 0;
        } // end of del[r]==0
        else
        {
            for (si=1; si<MAX_GF_EL_2; si++)
                sh[si] = gf9mul[del[r%MAX_GF_EL_2]][B[si-1]];
            sh[0]  = 0;
            for (si=0; si<MAX_GF_EL_2; si++)
            T[si] = gf9add[lam[si]][gf9addi[sh[si]]];
            if (2*L > r+rho-1)
            {
                for (si=0; si<MAX_GF_EL_2; si++)
                    lam[si] = T[si];
                for (si=MAX_GF_EL_2-1; si>0; si--)
                    B[si] = B[si-1];
                B[0] = 0;
            } // end of no change in L
            else
            {
                L    = r-L+rho;
                tmpi = inve2[del[r]];
                for (si=0; si<MAX_GF_EL_2; si++)
                    B[si] = gf9mul[tmpi][lam[si]];
                for (si=0; si<MAX_GF_EL_2; si++)
                    lam[si] = T[si];
            } // end of increase in L
        } // end of del[r]!=0
    } // end of (r <= 2*MAX_CORR_ERRORS)

    while (r<MAX_GF_EL_2)
    {
        r++;
        // calculate del[r]
        tmpi = 0;
        for (si=0; si<=(r<MAX_GF_EL_2 ? r: MAX_GF_EL_2-1); si++)
            tmpi = gf9add[tmpi][gf9mul[lam[si]][S[(r-si)%MAX_GF_EL_2]]];
        del[r%MAX_GF_EL_2] = tmpi;
        while (del[r%MAX_GF_EL_2] < 0)
            del[r%MAX_GF_EL_2] += NUM_GF_EL_2;
        S[r%MAX_GF_EL_2] = gf9add[S[r%MAX_GF_EL_2]][gf9addi[del[r%MAX_GF_EL_2]]];
    } // end of (r<=n)

    // correct transformed codeword
    for (si=0; si<MAX_GF_EL_2; si++)
        C[si] = gf9add[V[si]][gf9addi[S[si]]];

    // inverse transform corrected codeword
    tmpi = inve2[MAX_GF_EL_2%FIELD_CHAR];       // FIELD_CHAR is characteristic of field
                                // characteristic is # of elements in smallest sub-field
    for (si=0; si<MAX_GF_EL_2; si++)
    {
        tmpj = 0;
        for (di=0; di<MAX_GF_EL_2; di++)
        {
            texp = (si*di)%MAX_GF_EL_2;
            if (texp == 0)
                tmpk = 1;
            else
                tmpk = inva2[texp];
            tmpj = gf9add[tmpj][gf9mul[tmpk][C[di]]];
        } // end of inner inverse transform loop

        c[si] = gf9mul[tmpi][tmpj];
        while (c[si] < 0)
            c[si] += NUM_GF_EL_2;
    } // end of inverse transform

    // correct and count wrong symbols
    ec = 0;
    for (si=0; si<MAX_GF_EL_2; si++)
    {
        if (c[si] != codeword[si])
            {ec++; codeword[si]=c[si];}
    }

    return ec;
}

